import io
import sys

def hex_char_dump_test ():
    import zlx.bin

    d = zlx.bin.hex_char_dump(
            data = b'xxx\x01a\x02b\x03c\x04d\x05efghijklmnopqrstuvwxyz',
            data_offset = 3,
            data_length = 30,
            display_data_offset = 0x12307,
            display_row_offset = 0x12302,
            width = 13)
    print(d)
    assert d.strip() == '''
12302: .. .. .. ..  .. 01 61 02  62 03 63 04  64         .a.b.c.d
1230F: 05 65 66 67  68 69 6A 6B  6C 6D 6E 6F  70    .efghijklmnop
1231C: 71 72 73 74  75 76 77 78  79 .. .. ..  ..    qrstuvwxy
'''.strip()
    return

def wire_test ():
    import zlx.wire
    assert zlx.wire.u8.decode(io.BytesIO(b'abc')) == 0x61
    assert zlx.wire.u16le.decode(io.BytesIO(b'abc')) == 0x6261
    assert zlx.wire.u16be.decode(io.BytesIO(b'abc')) == 0x6162

    try:
        zlx.wire.u8.decode(io.BytesIO(b''))
    except zlx.wire.decode_error as e:
        assert 'truncated data' in e.args

    mm = {b'aragula': 0, b'barabu': 1, b'barz': 2, b'barabul': 3}
    assert zlx.wire.stream_decode_byte_seq_map(io.BytesIO(b'barabula'), mm) == 3

    assert zlx.wire.stream_decode_byte_seq_map(io.BytesIO(b'xbarabula'), mm, False) is None

    try:
        zlx.wire.stream_decode_byte_seq_map(io.BytesIO(b'xbarabula'), mm)
    except zlx.wire.decode_error as e:
        assert 'no match' in e.args[0]

    mc = zlx.wire.magic_codec('file_formats', b'MZ', b'\x7FELF', b'PNG')
    try:
        mc.decode(io.BytesIO(b'bla'))
    except zlx.wire.decode_error as e:
        assert 'no match' in e.args[0]

    f = io.BytesIO(b'MZAP')
    assert mc.decode(f) == b'MZ'
    assert f.read() == b'AP'

    field = zlx.wire.stream_record_field
    ABC = zlx.wire.stream_record_codec(
            '''
            ABC: # my ABC record
                magic: abc_magic # this is magic
                aaa: u16be
                u32le bbb
            ''',
            # field('magic', zlx.wire.magic_codec('abc', b'ABC\n', register = False)),
            # field('aaa', zlx.wire.u16be),
            # field('bbb', zlx.wire.u32le),
            field('ccc', zlx.wire.u8), abc_magic=zlx.wire.magic_codec('abc', b'ABC\n'))
    f = io.BytesIO(b'ABC\ndefghijk')
    o = ABC.decode(f)
    print('o = {!r}'.format(o))
    assert o.aaa == 0x6465
    assert o.bbb == 0x69686766
    assert o.ccc == 0x6A
    assert o.magic == b'ABC\n'

    s = zlx.wire.stream(b'0123456789', *zlx.wire.INT_CODECS)
    print('bla {}'.format(s.u32be[1]))
    assert s.u8[3] == 0x33
    return

def io_test ():
    import zlx.io
    b1 = io.BytesIO(b'0123456789abcdefghijklmnopqrstuvwxyz')
    b2 = io.BytesIO(b'ABCDEFGHIJKLMNOPQRSTUVWXYZ')
    cs = zlx.io.chunked_stream((zlx.io.chunk(b1, 1, 5), zlx.io.chunk(b2, 5, 5),
        zlx.io.chunk(b1, 3, 3)))
    r = cs.read(12)
    print(repr(r))
    assert r == b'12345FGHIJ34'

    ba = bytearray(b'012345')
    bav = zlx.io.ba_view(ba)
    assert len(bav) == 6
    bav.seek(-3, io.SEEK_END)
    assert bav.read(5) == b'345'
    return

def record_test ():
    import zlx.record
    import zlx.int
    P = zlx.record.make('Point', 'x y', validators=dict(int=zlx.int.u8_in_range))
    p = P(1, 2)
    print(repr(p))
    print(p.validate_x())

    # FBZ = zlx.record.make('FBZ', 'foo:u8 bar:u32le baz:u16be', field_repr=dict(a=zlx.int.hex, b=zlx.int.hex, c=zlx.int.hex))
    # f = zlx.bin.io_accessor(io.BytesIO(b'@abcdefghijk'))
    # fbz = FBZ.from_io_accessor(f, 1)[0]
    # print(repr(fbz))
    # assert fbz.foo == 0x61
    # assert fbz.bar == 0x65646362
    # assert fbz.baz == 0x6667
    return


def pe_test ():
    import zlx.bin
    import zlx.pe
    ba = zlx.wire.stream(b'MZ\0\0xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\x80\0\0\0')
    x = zlx.pe.parse_mz_header(ba)
    print(repr(x))
    assert x.magic == zlx.pe.MZ_MAGIC
    assert x.e_lfanew == 0x80

def map_pe (input_path, output_path):
    import zlx.io
    import zlx.bin
    import zlx.pe
    ba = zlx.wire.stream(zlx.io.bin_load(input_path))
    mzh = zlx.pe.parse_mz_header(ba)
    peh = zlx.pe.parse_pe_header(ba, offset = mzh.e_lfanew)
    #print(repr(mzh))
    #print(repr(peh))
    image = zlx.pe.map_parsed_pe(ba, peh)
    assert image[0:2] == b'MZ', 'image should start with MZ'
    hw_rva = image.find(b'Hello world!')
    assert hw_rva >= 0x1000
    assert (hw_rva & 0xFFF) < 0x100
    zlx.io.omsg('hello_msg offset: {:X}', hw_rva)
    zlx.io.bin_save(output_path, image)
    return

def windump_info (input_path):
    import zlx.windump
    import zlx.io
    with open(input_path, 'rb') as f:
        dh = zlx.windump.parse_header(f)
    print(repr(dh))
    return

def msf7_info (input_path):
    import zlx.msf7
    mr = zlx.msf7.reader(input_path)
    print(repr(mr.superblock))
    mr.load_dir()
    print('dir blocks: {!r}'.format(mr.dir_blocks))
    print('stream count: {}'.format(mr.stream_count))
    print('stream sizes: {!r}'.format(mr.stream_size_table))
    print('stream #1:\n{}'.format(zlx.bin.hex_char_dump(mr.streams[1])))

def xref_test ():
    import zlx.bin
    import zlx.wire
    x = zlx.bin.xref_scan(b'\xE8\x0B\0\0\0\xB8\x10\x10\0\0------X\xFA\xFF\xFF\xFF!',
            target = 0x1010,
            encoder = lambda v: zlx.wire.u32le.encode_to_bytes(v & 0xFFFFFFFF),
            base = 0x1000, rel_delta_range = range(4, 6))
    print('{!r}'.format(x))

def inc_build_no_test ():
    from zlx.build_number import inc as increment_build_number
    x = increment_build_number('ala\nbala\nBUILD =           	123; # gr\nxx;')
    print(repr(x))
    assert x == 'ala\nbala\nBUILD =           	124; # gr\nxx;\n'

if __name__ == '__main__':
    print(repr(sys.argv))
    if len(sys.argv) >= 2:
        if sys.argv[1] == 'map-pe':
            map_pe(sys.argv[2], sys.argv[3])
        elif sys.argv[1] == 'windump-info':
            windump_info(sys.argv[2])
        elif sys.argv[1] == 'msf7-info':
            msf7_info(sys.argv[2])
    else:
        var = None
        for var in sorted(filter(lambda n: n.endswith('_test'), globals())):
            print('calling test {!r}'.format(var))
            globals()[var]()

        #io_test()
        #record_test()
        #wire_test()
        #hex_char_dump_test()
        #pe_test()




from __future__ import absolute_import
import io
import re
import zlx.io

def _replace (match):
    return '{}{}'.format(match.group(1), int(match.group(2)) + 1)

def inc (source = None, path = None, pattern = 'BUILD =', save_path = None):
    if source is None and path is not None:
        with open(path, 'r') as f: source = f.read()
    if isinstance(pattern, str):
        pattern = re.compile(r'({}\s*)(\d+)'.format(pattern))
    out = io.StringIO()
    for line in source.splitlines():
        out.write(pattern.sub(_replace, line) + u'\n')
    content = out.getvalue()
    if save_path:
        with open(save_path, 'w') as f: f.write(content)
    return content




from __future__ import absolute_import
import struct
import io

import zlx.int
import zlx.record
import zlx.io

SEEK_SET = 0
SEEK_CUR = 1
SEEK_END = 2

class decode_error (RuntimeError): pass

PACK_FMT_DICT = {
    'u8': 'B',
    'i8': 'b',
    'u16le': '<H',
    'u16be': '>H',
    'i16le': '<h',
    'i16be': '>h',
    'u32le': '<I',
    'u32be': '>I',
    'i32le': '<i',
    'i32be': '>i',
    'u64le': '<Q',
    'u64be': '>Q',
    'i64le': '<q',
    'i64be': '>q',
}

CODEC_REGISTRY = {}

def default_desc (x):
    if isinstance(x, zlx.int.INT_TYPES):
        return '{}(0x{:X})'.format(x, x)
    return repr(x)

def register_codec (codec):
    if codec.name not in CODEC_REGISTRY:
        CODEC_REGISTRY[codec.name] = codec


class stream_codec (object):

    __slots__ = 'decode encode name desc'.split()

    def __init__ (self, name, decode, encode, desc = default_desc, register = True):
        self.name = name
        self.decode = decode
        self.encode = encode
        self.desc = desc
        if register: register_codec(self)

    def encode_to_bytes (self, value):
        f = io.BytesIO()
        self.encode(f, value)
        return f.getvalue()

    def decode_from_bytes (self, data):
        f = encoded_stream(data, self)
        return f.read()


def stream_decode_unpack (stream, pack_fmt, pack_len):
    data = stream.read(pack_len)
    try:
        return struct.unpack(pack_fmt, data)[0]
    except struct.error:
        raise decode_error('truncated data')

def stream_encode_pack (stream, value, pack_fmt):
    stream.write(struct.pack(pack_fmt, value))

def stream_decode_copy (stream, size, throw_on_no_match = True):
    v = stream.read(size)
    if len(v) != size and throw_on_no_match:
        raise decode_error('truncated data')
    return v

def stream_encode_copy (stream, value):
    stream.write(value)

def stream_decode_array (stream, codec, count):
    return tuple(codec.decode(stream) for i in range(count))

def dec_hex_int_desc (value):
    return '{0}(0x{0:X})'.format(value)

INT_CODECS = []
for codec_name in PACK_FMT_DICT:
    codec = stream_codec(
            name = codec_name,
            decode = lambda stream, pack_fmt=PACK_FMT_DICT[codec_name], pack_len=len(struct.pack(PACK_FMT_DICT[codec_name], 0)): stream_decode_unpack(stream, pack_fmt, pack_len),
            encode = lambda stream, value, pack_fmt=PACK_FMT_DICT[codec_name]: stream_encode_pack(stream, value, pack_fmt),
            desc = dec_hex_int_desc)
    globals()[codec_name] = codec
    INT_CODECS.append(codec)

def stream_decode_byte_seq_map (stream, byte_seq_map, throw_on_no_match = True):
    max_len = max(len(k) for k in byte_seq_map)
    data = stream.read(max_len)
    match = None
    for k in byte_seq_map:
        if data[0:len(k)] == k:
            if match is None or len(k) > len(match):
                match = k
    if match is None:
        if throw_on_no_match: raise decode_error('no match')
        match_len = 0
    else:
        match_len = len(match)
        if isinstance(byte_seq_map, dict):
            match = byte_seq_map[match]
    stream.seek(match_len - max_len, 1)
    return match

def magic_codec (name, *magic_list):
    if name.startswith('!'):
        name = name[1:]
        register = False
    else:
        register = True
    return stream_codec(
            name = name,
            decode = lambda stream, _map = magic_list: stream_decode_byte_seq_map(stream, _map),
            encode = stream_encode_copy,
            register = register)

stream_record_field = zlx.record.make('record_field', 'name codec desc')

#* stream_record_codec ******************************************************/
class stream_record_codec (object):
    __slots__ = 'name fields record_type'.split()
    def __init__ (self, name_or_spec, *fields, **kw):
        if '\n' in name_or_spec:
            name = None
            fl = []
            for line in name_or_spec.splitlines():
                if '#' in line: line = line[0:line.index('#')]
                line = line.strip()
                if not line: continue
                if name is None:
                    if not line.endswith(':'):
                        raise RuntimeError('bad record spec - expecting "name:\n"')
                    name = line[0:-1].strip()
                else:
                    if ':' in line:
                        fn, fc = (x.strip() for x in line.split(':', 1))
                    else:
                        fc, fn = line.split()
                    codec = kw[fc] if fc in kw else CODEC_REGISTRY[fc]
                    field = stream_record_field(fn, codec, codec.desc)
                    fl.append(field)
            fl.extend(fields)
            fields = fl
        else:
            name = name_or_spec

        self.name = name
        fields = tuple(fields)
        self.fields = fields
        #print(repr(tuple((f.name for f in fields))))
        self.record_type = zlx.record.make(name,
            fields = tuple(f.name for f in fields),
            field_repr = { f.name: f.desc or default_desc for f in fields })
        register_codec(self)

    def decode (self, stream):
        return self.record_type(**{f.name: f.codec.decode(stream) for f in self.fields})

    def encode (self, stream, value):
        for f in self.fields:
            f.codec.encode(stream, getattr(value, f.name))


#* encoded_stream ***********************************************************/
class encoded_stream (object):

    __slots = 'stream decode encode'.split()
    def __init__ (self, stream, codec):
        self.stream = stream
        self.decode = codec.decode
        self.encode = codec.encode

    def read (self, count = None):
        if count is not None:
            return [self.decode(self.stream) for i in range(count)]
        return self.decode(self.stream)

    def write (self, value):
        self.encode(self.stream, value)

    def __getitem__ (self, index):
        if isinstance(index, tuple):
            offset, count = index
            self.stream.seek(offset)
            return tuple(self.decode(stream) for i in range(count))
        else:
            self.stream.seek(index)
            return self.decode(self.stream)

    def __setitem__ (self, offset, value):
        self.stream.seek(offset)
        self.encode(self.stream, value)

    def __len__ (self):
        pos = self.stream.seek(0, SEEK_CUR)
        end = self.stream.seek(0, SEEK_END)
        self.stream.seek(pos, SEEK_SET)
        return end


#* stream *******************************************************************/
class stream (object):

    #__slots__ = 'stream codec_streams'.split()
    def __init__ (self, stream, *codec_list, **codec_map):
        if isinstance(stream, (bytes, bytearray)):
            stream = zlx.io.ba_view(stream)
        self.stream = stream
        if not codec_list and not codec_map:
            codec_map = CODEC_REGISTRY
        for codec in codec_list:
            if isinstance(codec, (str,)):
                codec = CODEC_REGISTRY[codec]
            self.add_codec(codec)
        for name, codec in codec_map.items():
            self.add_codec(codec, name)

    def add_codec (self, codec, name = None):
        if name is None: name = codec.name
        setattr(self, name, encoded_stream(self.stream, codec))

    def __len__ (self):
        pos = self.stream.seek(0, SEEK_CUR)
        end = self.stream.seek(0, SEEK_END)
        self.stream.seek(pos, SEEK_SET)
        return end

    def __getitem__ (self, index):
        if isinstance(index, tuple):
            offset, count = index
            self.stream.seek(offset)
            return self.stream.read(count)
        else:
            self.stream.seek(index)
            return self.stream.read(1)[0]

    def seekable (self):
        return self.stream.seekable()

    def seek (self, offset, whence = SEEK_SET):
        return self.stream.seek(offset, whence)

    def read (self, size = -1):
        return self.stream.read(size)




import sys

if sys.version_info[0] >= 3:
    INT_TYPES = (int,)
else:
    INT_TYPES = (int, long)

def pow2_check (n):
    return n > 0 and (n & (n - 1)) == 0

def pow2_round_down (n, p):
    assert pow2_check(p)
    return n & ~(p - 1)

def pow2_round_up (n, p):
    assert pow2_check(p)
    return (n + p - 1) & ~(p - 1)

def hex (n, prefix='0x', suffix=''):
    return '{}{:X}{}'.format(prefix, n, suffix)

def log2_ceil (n):
    p = 0
    while n > (1 << p): p += 1
    return p

def hex_items (l, sep = ', ', prefix='', suffix='', item_prefix='0x', item_suffix=''):
    return ''.join((prefix, sep.join(hex(n, prefix=item_prefix, suffix=item_suffix) for n in l), suffix))

def u8_hex (n):
    return '{:02X}'.format(n)

def u16_hex (n):
    return '{:04X}'.format(n)

def u32_hex (n):
    return '{:04X}'.format(n)

def u64_hex (n):
    return '{:04X}'.format(n)

def u8_in_range (n):
    return n >= 0 and n < 0x100

def u8_trunc (n):
    return n & 0xFF

def u8_add (a, b):
    return (a + b) & 0xFF



from __future__ import absolute_import
import sys
import io
import os
import zlx.int
import zlx.record

SEEK_SET = 0
SEEK_CUR = 1
SEEK_END = 2
SEEK_DATA = os.SEEK_DATA
SEEK_HOLE = os.SEEK_HOLE

def sfmt (fmt, *l, **kw): return fmt.format(*l, **kw)

if os.environ.get('DEBUG', ''):
    def dmsg (fmt, *l, **kw):
        sys.stdout.write((fmt + '\n').format(*l, **kw))
else:
    def dmsg (fmt, *l, **kw): pass

def omsg (fmt, *l, **kw):
    return sys.stdout.write((fmt + '\n').format(*l, **kw))

def emsg (fmt, *l, **kw):
    return sys.stderr.write((fmt + '\n').format(*l, **kw))

def bin_load (path):
    with open(path, 'rb') as f:
        return f.read()

def bin_save (path, content):
    with open(path, 'wb') as f:
        return f.write(content)

def txt_load (path):
    with open(path, 'r') as f:
        return f.read()

def txt_save (path, content):
    with open(path, 'w') as f:
        return f.write(content)

class chunk (object):
    __slots__ = 'stream offset size'.split()
    def __init__ (self, stream, offset, size):
        self.stream = stream
        self.offset = offset
        self.size = size

class chunked_stream (io.RawIOBase):

    def __init__ (self, io_chunks):
        self.io_chunks = tuple(io_chunks)
        self.chunk_pos = []
        pos = 0
        for c in self.io_chunks:
            self.chunk_pos.append(pos)
            pos += c.size
        self.size = pos
        self.chunk_pos.append(pos)
        self.pos = 0

    def seekable (self):
        return True

    def seek (self, offset, whence=SEEK_SET):
        if whence == SEEK_SET: pass
        elif whence == SEEK_CUR: offset += self.pos
        elif whence == SEEK_END: offset += self.size
        else: raise ValueError('unsupported whence {}'.format(whence))
        if offset < 0: raise ValueError('negative offset')
        self.pos = offset
        return offset

    def offset_to_chunk_index (self, offset):
        a, b = 0, len(self.io_chunks) - 1
        while a <= b:
            c = (a + b) // 2
            if offset >= self.chunk_pos[c]:
                if offset < self.chunk_pos[c + 1]:
                    return c
                a = c + 1
            else:
                b = c - 1
        return None

    def readinto (self, b):
        size = len(b)
        #print('readinto pos={} size={}'.format(self.pos, size))
        out_ofs = 0
        while out_ofs < size:
            cx = self.offset_to_chunk_index(self.pos)
            if cx is None: break
            offset_in_chunk = self.pos - self.chunk_pos[cx]
            cplen = min(size - out_ofs, self.io_chunks[cx].size - offset_in_chunk)
            #print('cx={} oic={} seekpos={} cplen={}'.format(cx, self.io_chunks[cx].offset + offset_in_chunk, self.io_chunks[cx].offset + offset_in_chunk, cplen))
            self.io_chunks[cx].stream.seek(self.io_chunks[cx].offset + offset_in_chunk)
            #print('before data={!r}'.format(b[out_ofs:out_ofs + cplen]))
            data = self.io_chunks[cx].stream.read(cplen)
            n = len(data)
            b[out_ofs:out_ofs + n] = data
            #print('n={} data={!r}'.format(n, b[out_ofs:out_ofs + cplen]))
            out_ofs += n
            self.pos += n
            if n != cplen: break
        return out_ofs

class ba_view (io.RawIOBase):
    '''
    Creates a stream backed by an existing bytearray-like object.
    '''
    __slots__ = 'ba pos'.split()
    def __init__ (self, ba):
        self.ba = ba
        self.pos = 0
    def seekable (self): return True
    def seek (self, offset, whence = SEEK_SET):
        if whence == SEEK_SET: pass
        elif whence == SEEK_CUR: offset += self.pos
        elif whence == SEEK_END: offset += len(self.ba)
        else: raise ValueError('unsupported whence {}'.format(whence))
        if offset < 0: raise ValueError('negative offset')
        self.pos = offset
        return offset
    def readinto (self, b):
        cplen = min(len(b), len(self.ba) - self.pos)
        if cplen <= 0: return None
        b[0:cplen] = self.ba[self.pos : self.pos + cplen]
        return cplen
    def __len__ (self):
        return len(self.ba)

#/* stream_cache *************************************************************/
SCK_UNCACHED = 0
SCK_CACHED = 1
SCK_HOLE = 2

uncached_data_block = zlx.record.make('uncached_data_block', 'offset size')
uncached_data_block.get_size = lambda x: x.size
uncached_data_block.kind = SCK_UNCACHED
uncached_data_block.desc = lambda x: sfmt('uncached(0x{:X},0x{:X})', x.offset, x.size)

cached_data_block = zlx.record.make('cached_data_block', 'offset data')
cached_data_block.get_size = lambda x: len(x.data)
cached_data_block.kind = SCK_CACHED
cached_data_block.desc = lambda x: sfmt('cached(0x{:X},0x{:X},{!r})', x.offset, len(x.data), bytes(x.data[0:4]))

hole_block = zlx.record.make('hole_block', 'offset size')
hole_block.get_size = lambda x: x.size
hole_block.kind = SCK_HOLE
hole_block.desc = lambda x: sfmt('hole(0x{:X},0x{:X})', x.offset, x.size) if x.size else sfmt('end(0x{:X})', x.offset)

class stream_cache (object):

    def __init__ (self, stream, align = 4096, assume_size = None):

        self.stream = stream

        self.seekable = False
        self.blocks = []

        if assume_size is not None:
            self.seekable = stream.seekable()
            end = assume_size
        elif stream.seekable():
            try:
                self.pos = stream.seek(0, SEEK_CUR)
                end = stream.seek(0, SEEK_END)
                stream.seek(self.pos, SEEK_CUR)
                self.seekable = True
            except:
                pass
        if self.seekable:
            self.blocks.append(uncached_data_block(0, end))
            self.blocks.append(hole_block(end, 0))
            assert zlx.int.pow2_check(align), "alignment must be a power of 2"
            self.align = align # alignment for offsets / sizes when doing I/O
        else:
            self.blocks.append(hole_block(0, 0))
            self.align = 1

    def __repr__ (self):
        return sfmt('stream_cache(stream={!r}, seekable={!r}, blocks=[{}])', self.stream, self.seekable, ', '.join([x.desc() for x in self.blocks]))

    def locate_block (self, offset):
        for i in range(len(self.blocks)):
            b = self.blocks[i]
            if offset >= b.offset and offset - b.offset < b.get_size():
                return i, b
        return len(self.blocks) - 1, self.blocks[-1]

    def get_part (self, offset, size):
        '''
        returns information from cache about data starting with given offset.
        The information returned may describe a smaller portion than the requested size
        but never more. The caller must call again to get information about the
        remaining data
        '''
        bx, b = self.locate_block(offset)
        if b.kind == SCK_UNCACHED:
            assert b.offset <= offset and offset - b.offset < b.size
            return uncached_data_block(offset, min(size, b.offset + b.size - offset))
        elif b.kind == SCK_CACHED:
            b_size = b.get_size()
            assert b.offset <= offset and offset - b.offset < b_size
            n = min(size, b.offset + b_size - offset)
            o = offset - b.offset
            return cached_data_block(offset, b.data[o : o + n])
        elif b.kind == SCK_HOLE:
            assert b.offset <= offset and offset - b.offset < b.size
            return uncached_data_block(offset, min(size, b.offset + b.size - offset))
        else:
            return b

    def get (self, offset, size):
        '''
        returns a list of blocks that describe the given range as returned
        byte get_part
        '''
        a = []
        while size:
            blk = self.get_part(offset, size)
            a.append(blk)
            size -= blk.get_size() or size
        return a

    def _seek (self, offset):
        if self.seekable:
            assert offset >= 0, 'cannot seek to negative offsets'
            self.stream.seek(offset, SEEK_SET)
        else:
            if offset != self.pos:
                raise RuntimeError("unseekable cannot change pos from {} to {}".format(self.pos, offset))

    def _load (self, offset, size):
        o = zlx.int.pow2_round_down(offset, self.align)
        e = zlx.int.pow2_round_up(offset + size, self.align)
        dmsg('load o={:X} e={:X}', o, e)
        self._seek(o)
        while o < e:
            data = self.stream.read(e - o)
            if not data:
                self._update_no_data(o)
                break
            self._update_data(o, data)
            o += len(data)

    def _merge_left (self, bx):
        if bx == 0 or bx >= len(self.blocks): return
        l = self.blocks[bx - 1]
        r = self.blocks[bx]
        if l.kind != r.kind: return
        assert l.offset + l.get_size() == r.offset, 'non-contiguous blocks'
        if l.kind == SCK_CACHED:
            l.data[len(l.data):] = r.data
            del self.blocks[bx]

    def _merge_around (self, bx, count = 1):
        self._merge_left(bx)
        self._merge_left(bx + count)

    def _update_data (self, offset, data):
        dmsg('updating o=0x{:X} len=0x{:X}', offset, len(data))
        while data:
            bx, b = self.locate_block(offset)
            dmsg('ofs=0x{:X} len=0x{:X}. got block: {}', offset, len(data), b.desc())
            if b.kind == SCK_HOLE:
                if offset > b.offset:
                    self.blocks.insert(bx, uncached_data_block(b.offset, offset - b.offset) )
                    bx += 1
                self.blocks.insert(bx, cached_data_block(offset, bytearray(data)))
                self._merge_left(bx)
                b.offset = offset + len(data)
                return
            elif b.kind == SCK_UNCACHED:
                new_blocks = []
                b_end = b.offset + b.size
                if b.offset < offset:
                    new_blocks.append(uncached_data_block(b.offset, offset - b.offset))
                nb_len = min(b_end - offset, len(data))
                new_blocks.append(cached_data_block(offset, bytearray(data[0: nb_len])))
                data_end = offset + len(data)
                if data_end < b_end:
                    new_blocks.append(uncached_data_block(data_end, b_end - data_end))
                self.blocks[bx : bx + 1] = new_blocks
                self._merge_around(bx, len(new_blocks))
                offset += nb_len
                data = data[nb_len:]
            elif b.kind == SCK_CACHED:
                b_end = b.offset + len(b.data)
                update_len = min(b_end - offset, len(data))
                b.data[offset - b.offset : offset - b.offset + update_len] = data[0 : update_len]
                offset += update_len
                data = data[update_len:]
            else:
                raise sfmt("huh? {!r}", b)

        pass

    def _split_block (self, bx, offset):
        '''
        splits a block that has size (cached, uncached, hole) and returns the index and
        the block that starts with the given offset.
        '''
        blk = self.blocks[bx]
        assert blk.kind in (SCK_CACHED, SCK_UNCACHED, SCK_HOLE)
        assert blk.offset <= offset
        assert blk.offset + blk.get_size() > offset
        if blk.offset == offset: return bx, blk
        if blk.kind == SCK_CACHED:
            nblk = cached_data_block(offset, data = bytearray(blk.data[offset - blk.offset:]))
            blk.data[offset - blk.offset:] = b''
        else:
            nblk = blk.__class__(offset = offset, size = blk.offset + blk.size - offset)
        self.blocks.insert(bx + 1, nblk)
        return bx + 1, nblk

    def _discard_contiguous_data_blocks (self, bx):
        '''
        deletes all blocks from given index as long as they refer to data 
        (cached/uncached) and updates the next non-data block (fix offset/size).
        '''
        offset = self.blocks[bx].offset
        while self.blocks[bx].kind in (SCK_CACHED, SCK_UNCACHED):
            del self.blocks[bx]
        assert self.blocks[bx].kind in (SCK_HOLE, )
        if self.blocks[bx].kind == SCK_HOLE:
            self.blocks[bx].size += offset - self.blocks[bx].offset
        self.blocks[bx].offset = offset
        self._merge_left(bx)

    def _update_no_data (self, offset):
        bx, blk = self.locate_block(offset)
        if blk.kind in (SCK_CACHED, SCK_UNCACHED):
            bx, blk = self._split_block(bx, offset)
            self._discard_contiguous_data_blocks(self, bx)
        pass




import struct
import sys
import zlx.int
import zlx.wire

if sys.version_info[0] >= 3:
    from io import StringIO
else:
    from StringIO import StringIO

NO_OFFSET = -1
PLAIN_OFFSET = 128
GROUP_4_OFFSET = 4

NO_HEX = None

NO_CHARS = None
PRINTABLE_ASCII_CHARS_MAP = tuple('.' if n < 0x20 or n >= 0x7F else chr(n) for n in range(256))

OFFSET_8HEX = lambda n: '{:08X}: '.format(n)
OFFSET_HIDE = lambda n: ''

def byte_as_printable_ascii_char_printer (data, offset):
    return (PRINTABLE_ASCII_CHARS_MAP[data[offset]], 1)

def hex_char_dump (
        data,
        data_offset = 0,
        data_length = None,
        display_data_offset = None,
        display_row_offset = None,
        display_row_count = None,
        offset_printer = None,
        char_printer = byte_as_printable_ascii_char_printer,
        width = 0x10,
        line_prefix = '',
        line_suffix = '\n',
        offset_suffix = ': ',
        mod_sep = { 1: ' ', 4: '  ' },
        hex_char_sep = '    ',
        no_data_pad = '..',
        no_char_pad = ' ',
        ):

    if isinstance(data, (bytes, bytearray)):
        data = zlx.wire.stream(data, 'u8').u8

    if data_length is None:
        data_length = len(data) - data_offset

    if display_data_offset is None:
        display_data_offset = data_offset
    if display_row_offset is None:
        display_row_offset = display_data_offset - display_data_offset % width
    if display_row_count is None:
        items_in_first_row = width - (display_data_offset - display_row_offset)
        display_row_count = 1
        if data_length > items_in_first_row:
            display_row_count += (data_length - items_in_first_row + width - 1) // width

    if offset_printer is None:
        last_offset = display_data_offset + (display_row_count - 1) * width
        offset_digits = (zlx.int.log2_ceil(last_offset) + 3) // 4

        offset_printer = lambda x, fmt='{{:0{}X}}: '.format(offset_digits): fmt.format(x)

    offset_delta = data_offset - display_data_offset
    o = StringIO()
    for r in range(display_row_count):
        o.write(offset_printer(display_row_offset + r * width))
        for c in range(width):
            offset = display_row_offset + r * width + c + offset_delta
            if c > 0:
                for b in sorted(mod_sep.keys(), reverse=True):
                    if c % b == 0:
                        o.write(mod_sep[b])
                        break
            if offset < data_offset or offset >= data_offset + data_length:
                o.write(no_data_pad)
            else:
                o.write('{:02X}'.format(data[offset]))
        o.write(hex_char_sep)
        inc = 1
        for c in range(width):
            inc -= 1
            if inc > 0: next
            offset = display_row_offset + r * width + c + offset_delta
            if offset < data_offset or offset >= data_offset + data_length:
                o.write(no_char_pad)
            else:
                s, inc = char_printer(data, offset)
                o.write(s)
        o.write('\n')

    return o.getvalue()

xref_collection = zlx.record.make('xrefs', 'absolute relative')

def xref_scan (data, target, encoder, base = 0, 
        start_offset = 0, end_offset = None,
        abs_ref = True, rel_delta_range = range(0, 1)):
    if end_offset is None: end_offset = len(data)
    xrefs = xref_collection([], {})
    if abs_ref:
        encoded_target = encoder(target)
        offset = start_offset
        while True:
            offset = data.find(encoded_target, offset, end_offset)
            if offset < 0: break
            xrefs.absolute.append(offset)
            offset += 1
    if rel_delta_range:
        for offset in range(start_offset, end_offset):
            for delta in rel_delta_range:
                encoded_target = encoder(target - base - delta - offset)
                print('checking for {!r} at offset {!r} where {!r} is'.format(encoded_target, offset, data[offset : offset + len(encoded_target)]))
                if data[offset : offset + len(encoded_target)] == encoded_target:
                    if delta not in xrefs.relative:
                        xrefs.relative[delta] = []
                    xrefs.relative[delta].append(offset)
    return xrefs




from setuptools import setup
import sys
import os
sys.path.append(os.path.dirname(os.path.realpath(__file__)))
import zlx

ZLX_DESC = '''
# Zalmoxis

Zalmoxis is a python module for text and binary manipulation.
zlx.io submodule offers convenient wrappers for streams and byte-array
that encode/decode items.

'''

setup(name='zlx',
      version = zlx.VER_STR,
      description = 'Zalmoxis - module for text and binary manipulation',
      long_description = ZLX_DESC,
      long_description_content_type = "text/markdown",
      url = 'https://gitlab.com/icostin/zlx-py',
      author = 'Costin Ionescu',
      author_email = 'costin.ionescu@gmail.com',
      license = 'MIT',
      packages = ['zlx'],
      zip_safe = False,
      entry_points = {
          'console_scripts': ['zlx=zlx.cmd_line:entry'],
      },
      classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
      )




import argparse
import sys
import traceback
import zlx.record
import zlx.io

sfmt = zlx.io.sfmt
omsg = zlx.io.omsg
emsg = zlx.io.emsg
dmsg = zlx.io.dmsg

def cmd_help (req):
    req.ap.print_help()

def cmd_map_pe (req):
    import zlx.pe
    n = 0
    for input_path in req.FILE:
        output_path = req.out_spec.format(path=input_path, index=n)
        image = zlx.pe.map_pe_from_path(input_path, req.page_size)
        zlx.io.bin_save(output_path, image)
        n += 1

def cmd_msf7_info (req):
    import zlx.msf7
    for input_path in req.FILE:
        try:
            mr = zlx.msf7.reader(input_path)
            omsg('{!r}:', input_path)
            omsg(' superblock:')
            omsg('  magic:                  {!r}', mr.superblock.magic)
            omsg('  block size:             {}', mr.superblock.block_size)
            omsg('  free block map block:   {}', mr.superblock.free_block_map_block)
            omsg('  block count:            {}', mr.superblock.block_count)
            omsg('  dir size:               {}', mr.superblock.dir_size)
            omsg('  something:              {}', mr.superblock.suttin)
            omsg('  dir block map block:    {}', mr.superblock.dir_block_map_block)

            omsg(' directory:')
            mr.load_dir()
            omsg('  stream count:           {}', mr.stream_count)

            omsg(' streams:')
            for i in range(mr.stream_count):
                omsg('  {:03}: size={:<7}', i, mr.stream_size_table[i])
        except Exception as e:
            emsg('error processing file {!r}', input_path)
            raise

def cmd_inc_build (req):
    from zlx.build_number import inc
    for path in req.FILE:
        inc(path = path, save_path = path, pattern = req.prefix)

def cmd_test_stream_cache (req):
    with open(req.FILE, 'rb') as f:
        sc = zlx.io.stream_cache(f,
                align = req.alignment,
                assume_size = req.assume_size)
        for c in req.commands:
            dmsg('*** {!r}', sc)
            verb, *args = c.split(':')
            if verb == 'get':
                ofs, size = (int(x) for x in args)
                dmsg('--- get(offset={}, size={})', ofs, size)
                blk_list = sc.get(ofs, size)
                dmsg('==> block: {!r}', blk_list)
            elif verb == 'load':
                ofs, size = (int(x) for x in args)
                dmsg('--- load(offset={}, size={})', ofs, size)
                sc._load(ofs, size)
            else:
                raise RuntimeError(sfmt('unsupported verb {!r}', verb))
        dmsg('*** {!r}', sc)

    pass

def main (args):
    ap = argparse.ArgumentParser(
            description='tool to process binary and text data')
    ap.add_argument('-V', '--version', action='version', version='zlx {}'.format(zlx.VER_STR))
    ap.add_argument('-v', '--verbose', help='be verbose',
            action='store_true', default=False)

    sp = ap.add_subparsers(title='subcommands', dest='cmd')

    p = sp.add_parser('help')
    p = sp.add_parser('msf7-info',
            help='provides information about MSF v7 files')
    p.add_argument('FILE', nargs='*', help='file(s) to process')

    p = sp.add_parser('map-pe', help='creates an image of the mapped PE file')
    p.add_argument('FILE', nargs='*', help='input file(s)')
    p.add_argument('-o', '--out', dest='out_spec', default='{path}.img')
    p.add_argument('-p', '--page-size', dest='page_size', type=int, default=4096)

    p = sp.add_parser('inc-build',
            help = 'increments build number in a given file')
    p.add_argument('-p', '--prefix', dest = 'prefix', help='prefix string for the build number', default = 'BUILD =')
    p.add_argument('FILE', nargs='*', help='file(s)')

    p = sp.add_parser('test-stream-cache',
            help = 'test drives zlx.io.stream_cache')
    p.add_argument('FILE',
            help = 'input file')
    p.add_argument('-a', '--alignment',
            type = int,
            help = 'alignment for load commands',
            default = 4096)
    p.add_argument('-z', '--assume-size',
            type = int,
            help='init caching to assume the given size',
            default = None)
    p.add_argument('commands',
            nargs = '*',
            help = '"get:<offset>:<size>" or "load:<offset>:<size>"')

    req = ap.parse_args(args[1:])
    if req.verbose:
        print('command line: {!r}'.format(req))

    if req.cmd is None: req.cmd = 'help'
    dmsg('parsed command line: {}', req)

    req.ap = ap

    globals()['cmd_' + req.cmd.replace('-', '_')](req)

def entry ():
    main(sys.argv)




import zlx.record
import zlx.bin
import zlx.int

DUMP32_MAGIC = b'PAGEDUMP'
DUMP64_MAGIC = b'PAGEDU64'

Header = zlx.record.make('WinDumpHeader', '''
    magic
    ver_major
    ver_minor
    dir_table_base
    pfn_database
    loaded_module_list
    active_process_head
    machine_image_type
    cpu_count
    bug_check_code
    bug_check_params
    debugger_data_block
    phys_mem_blocks
    exc
    dump_type
    raw_system_up_time
    raw_system_time
    ''',
    field_repr = dict(
        dir_table_base = zlx.int.hex,
        pfn_database = zlx.int.hex,
        loaded_module_list = zlx.int.hex,
        active_process_head = zlx.int.hex,
        machine_image_type = zlx.int.hex,
        bug_check_code = zlx.int.hex,
        bug_check_params = lambda x: zlx.int.hex_items(x, sep=', ', prefix='(', suffix=')'),
        debugger_data_block = zlx.int.hex,
        ))

def parse_header (data):
    if isinstance(data, zlx.bin.io_accessor): ba = data
    else: ba = zlx.bin.io_accessor(data)
    h = Header(magic = ba[0, 8])
    if h.magic == DUMP32_MAGIC:
        h.ver_major = ba.u32le[0x08]
        h.ver_minor = ba.u32le[0x0C]
    elif h.magic == DUMP64_MAGIC:
        h.ver_major = ba.u32le[0x08]
        h.ver_minor = ba.u32le[0x0C]
        h.dir_table_base = ba.u64le[0x10]
        h.pfn_database = ba.u64le[0x18]
        h.loaded_module_list = ba.u64le[0x20]
        h.active_process_head = ba.u64le[0x28]
        h.machine_image_type = ba.u32le[0x30]
        h.cpu_count = ba.u32le[0x34]
        h.bug_check_code = ba.u32le[0x38]
        h.bug_check_params = ba.u64le[0x40, 4]
        h.debugger_data_block = ba.u64le[0x80]
    return h

